@@ -14,7 +14,7 @@ use Graph::AdjacencyMap qw(:flags :fields);
 
 use vars qw($VERSION);
 
-$VERSION = '0.77';
+$VERSION = '0.79';
 
 require 5.006; # Weak references are absolutely required.
 
@@ -440,8 +440,9 @@ sub _add_edge {
 	    push @e, $V->[ _s ]->{ $v };
 	}
     } else {
+	my $h = $g->[ _V ]->_is_HYPER;
 	for my $v ( @_ ) {
-	    my @v = ref $v eq 'ARRAY' ? @$v : $v;
+	    my @v = ref $v eq 'ARRAY' && $h ? @$v : $v;
 	    $g->add_vertex( @v ) unless $V->has_path( @v );
 	    push @e, $V->_get_path_id( @v );
 	}
@@ -486,8 +487,9 @@ sub _vertex_ids {
 	    push @e, $V->[ _s ]->{ $v };
 	}
     } else {
+	my $h = $g->[ _V ]->_is_HYPER;
 	for my $v ( @_ ) {
-	    my @v = ref $v eq 'ARRAY' ? @$v : $v;
+	    my @v = ref $v eq 'ARRAY' && $h ? @$v : $v;
 	    return () unless $V->has_path( @v );
 	    push @e, $V->_get_path_id( @v );
 	}
@@ -701,8 +703,9 @@ sub vertices_at {
     my @v;
     while (my ($i, $v) = each %{ $Vi }) {
 	my %i;
+	my $h = $V->[_f ] & _HYPER;
 	@i{ @i } = @i if @i; # @todo: nonuniq hyper vertices?
-	for my $u ((ref($v) eq 'ARRAY') ? @$v : $v) {
+	for my $u (ref $v eq 'ARRAY' && $h ? @$v : $v) {
 	    my $j = exists $v{ $u } ? $v{ $u } : ( $v{ $u } = $i );
 	    if (defined $j && exists $i{ $j }) {
 		delete $i{ $j };
@@ -723,10 +726,9 @@ sub _edges_at {
     my @e;
     my $en = 0;
     my %ev;
-    for my $v ( ($V->[_f ] & _HYPER) ? $g->vertices_at( @_ ) : @_ ) {
-#use Carp;
-#confess("v: $v, ref=".ref($v)) unless ref($v) eq 'ARRAY'; # Added by bsb
-	my $vi = $V->_get_path_id( (ref($v) eq 'ARRAY') ? @$v : $v );
+    my $h = $V->[_f ] & _HYPER;
+    for my $v ( $h ? $g->vertices_at( @_ ) : @_ ) {
+	my $vi = $V->_get_path_id( ref $v eq 'ARRAY' && $h ? @$v : $v );
 	next unless defined $vi;
 	my $Ei = $E->_ids;
 	while (my ($ei, $ev) = each %{ $Ei }) {
@@ -753,8 +755,9 @@ sub _edges_from {
     my $o = $E->[ _f ] & _UNORD;
     my $en = 0;
     my %ev;
-    for my $v ( ($V->[_f ] & _HYPER) ? $g->vertices_at( @_ ) : @_ ) {
-	my $vi = $V->_get_path_id( (ref($v) eq 'ARRAY') ? @$v : $v );
+    my $h = $V->[_f ] & _HYPER;
+    for my $v ( $h ? $g->vertices_at( @_ ) : @_ ) {
+	my $vi = $V->_get_path_id( ref $v eq 'ARRAY' && $h ? @$v : $v );
 	next unless defined $vi;
 	my $Ei = $E->_ids;
 	if (wantarray) {
@@ -804,8 +807,9 @@ sub _edges_to {
     my $o = $E->[ _f ] & _UNORD;
     my $en = 0;
     my %ev;
-    for my $v ( ($V->[_f ] & _HYPER) ? $g->vertices_at( @_ ) : @_ ) {
-	my $vi = $V->_get_path_id( (ref($v) eq 'ARRAY') ? @$v : $v );
+    my $h = $V->[_f ] & _HYPER;
+    for my $v ( $h ? $g->vertices_at( @_ ) : @_ ) {
+	my $vi = $V->_get_path_id( ref $v eq 'ARRAY' && $h ? @$v : $v );
 	next unless defined $vi;
 	my $Ei = $E->_ids;
 	if (wantarray) {
@@ -2516,6 +2520,55 @@ sub directed_copy {
 
 *directed_copy_graph = \&directed_copy;
 
+###
+# Cache or not.
+#
+
+my %_cache_type =
+    (
+     'connectivity'        => '_ccc',
+     'strong_connectivity' => '_scc',
+     'biconnectivity'      => '_bcc',
+     'SPT_Dijkstra'        => '_spt_di',
+     'SPT_Bellman_Ford'    => '_spt_bf',
+    );
+
+sub _clear_cache {
+    my ($g, $type) = @_;
+    my $c = $_cache_type{$type};
+    if (defined $c) {
+	$g->delete_graph_attribute($c);
+    } else {
+	Carp::croak("Graph: unknown cache type '$type'");
+    }
+}
+
+sub connectivity_clear_cache {
+    my $g = shift;
+    _clear_cache($g, 'connectivity');
+}
+
+sub strong_connectivity_clear_cache {
+    my $g = shift;
+    _clear_cache($g, 'strong_connectivity');
+}
+
+sub biconnectivity_clear_cache {
+    my $g = shift;
+    _clear_cache($g, 'biconnectivity');
+}
+
+sub SPT_Dijkstra_clear_cache {
+    my $g = shift;
+    _clear_cache($g, 'SPT_Dijkstra');
+    $g->delete_graph_attribute('SPT_Dijkstra_first_root');
+}
+
+sub SPT_Bellman_Ford_clear_cache {
+    my $g = shift;
+    _clear_cache($g, 'SPT_Bellman_Ford');
+}
+
 ###
 # Connected components.
 #
@@ -2736,13 +2789,14 @@ sub _strongly_connected_components {
 sub strongly_connected_components {
     my $g = shift;
     $g->expect_directed;
-    $g->_strongly_connected_components;
+    $g->_strongly_connected_components(@_);
 }
 
 sub strongly_connected_component_by_vertex {
-    my ($g, $v) = @_;
+    my $g = shift;
+    my $v = shift;
     $g->expect_directed;
-    my @scc = $g->_strongly_connected_components( next_alphabetic => 1);
+    my @scc = $g->_strongly_connected_components( next_alphabetic => 1, @_ );
     for (my $i = 0; $i <= $#scc; $i++) {
 	for (my $j = 0; $j <= $#{ $scc[$i] }; $j++) {
 	    return $i if $scc[$i]->[$j] eq $v;
@@ -2752,16 +2806,17 @@ sub strongly_connected_component_by_vertex {
 }
 
 sub strongly_connected_component_by_index {
-    my ($g, $i) = @_;
+    my $g = shift;
+    my $i = shift;
     $g->expect_directed;
-    my $c = ( $g->_strongly_connected_components )[ $i ];
+    my $c = ( $g->_strongly_connected_components(@_) )[ $i ];
     return defined $c ? @{ $c } : ();
 }
 
 sub same_strongly_connected_components {
     my $g = shift;
     $g->expect_directed;
-    my @scc = $g->_strongly_connected_components( next_alphabetic => 1);
+    my @scc = $g->_strongly_connected_components( next_alphabetic => 1, @_ );
     my @i;
     while (@_) {
 	my $v = shift;
@@ -2931,10 +2986,12 @@ sub biconnectivity {
 	}
 	$I{ $r } = 1;
 	my %U;
+	my %S; # Self-loops.
 	for my $e ($g->edges) {
 	    my ($u, $v) = @$e;
 	    $U{ $u }{ $v } = 0;
 	    $U{ $v }{ $u } = 0;
+	    $S{ $u } = 1 if $u eq $v;
 	}
 	my $i = 1;
 	my $v = $r;
@@ -2948,7 +3005,6 @@ sub biconnectivity {
 	# print "r : $r\n";
 
 	my %T; @T{ @V } = @V;
-	my %S; # Self-loops.
 
 	for my $w (@V) {
 	    my @s = $g->successors( $w );
@@ -3084,13 +3140,13 @@ sub biconnectivity {
 
 sub is_biconnected {
     my $g = shift;
-    my ($ap, $bc) = ($g->biconnectivity)[0, 1];
+    my ($ap, $bc) = ($g->biconnectivity(@_))[0, 1];
     return defined $ap ? @$ap == 0 && $g->vertices >= 3 : undef;
 }
 
 sub is_edge_connected {
     my $g = shift;
-    my ($br) = ($g->biconnectivity)[2];
+    my ($br) = ($g->biconnectivity(@_))[2];
     return defined $br ? @$br == 0 && $g->edges : undef;
 }
 
@@ -3102,7 +3158,7 @@ sub is_edge_separable {
 
 sub articulation_points {
     my $g = shift;
-    my ($ap) = ($g->biconnectivity)[0];
+    my ($ap) = ($g->biconnectivity(@_))[0];
     return defined $ap ? @$ap : ();
 }
 
@@ -3110,26 +3166,28 @@ sub articulation_points {
 
 sub biconnected_components {
     my $g = shift;
-    my ($bc) = ($g->biconnectivity)[1];
+    my ($bc) = ($g->biconnectivity(@_))[1];
     return defined $bc ? @$bc : ();
 }
 
 sub biconnected_component_by_index {
-    my ($g, $i) = @_;
-    my ($bc) = ($g->biconnectivity)[1];
+    my $g = shift;
+    my $i = shift;
+    my ($bc) = ($g->biconnectivity(@_))[1];
     return defined $bc ? $bc->[ $i ] : undef;
 }
 
 sub biconnected_component_by_vertex {
-    my ($g, $v) = @_;
-    my ($v2bc) = ($g->biconnectivity)[3];
+    my $g = shift;
+    my $v = shift;
+    my ($v2bc) = ($g->biconnectivity(@_))[3];
     return defined $v2bc->{ $v } ? keys %{ $v2bc->{ $v } } : ();
 }
 
 sub same_biconnected_components {
     my $g = shift;
     my $u = shift;
-    my @u = $g->biconnected_component_by_vertex($u);
+    my @u = $g->biconnected_component_by_vertex($u, @_);
     return 0 unless @u;
     my %ubc; @ubc{ @u } = ();
     while (@_) {
@@ -3152,7 +3210,7 @@ sub same_biconnected_components {
 
 sub biconnected_graph {
     my ($g, %opt) = @_;
-    my ($bc, $v2bc) = ($g->biconnectivity)[1, 3];
+    my ($bc, $v2bc) = ($g->biconnectivity, %opt)[1, 3];
     my $bcg = Graph::Undirected->new;
     my $sc_cb =
 	exists $opt{super_component} ?
@@ -3186,7 +3244,7 @@ sub biconnected_graph {
 
 sub bridges {
     my $g = shift;
-    my ($br) = ($g->biconnectivity)[2];
+    my ($br) = ($g->biconnectivity(@_))[2];
     return defined $br ? @$br : ();
 }
 
@@ -3225,14 +3283,14 @@ sub SPT_Dijkstra {
     my $spt_di = $g->get_graph_attribute('_spt_di');
     unless (defined $spt_di && exists $spt_di->{ $first_root } && $spt_di->{ $first_root }->[ 0 ] == $g->[ _G ]) {
 	my %etc;
-	my $sptg = $g->_heap_walk($g->new(), \&_SPT_add, \%etc, %opt);
+	my $sptg = $g->_heap_walk($g->new, \&_SPT_add, \%etc, %opt);
 	$spt_di->{ $first_root } = [ $g->[ _G ], $sptg ];
 	$g->set_graph_attribute('_spt_di', $spt_di);
     }
 
     my $spt = $spt_di->{ $first_root }->[ 1 ];
 
-    $spt->set_graph_attribute('SPT_Dijsktra_root', $first_root);
+    $spt->set_graph_attribute('SPT_Dijkstra_root', $first_root);
 
     return $spt;
 }
@@ -3333,7 +3391,7 @@ sub SPT_Bellman_Ford {
 	my ($p, $d) =
 	    $g->_SPT_Bellman_Ford($opt, $unseenh, $unseena,
 				  $r, $next, $code, $attr);
-	my $h = $g->new();
+	my $h = $g->new;
 	for my $v (keys %$p) {
 	    my $u = $p->{ $v };
 	    $h->add_edge( $u, $v );
@@ -3602,6 +3660,8 @@ sub center_vertices {
     return @c;
 }
 
+*centre_vertices = \&center_vertices;
+
 sub average_path_length {
     my $g = shift;
     my @A = @_;
